package ru.landgrafhomyak.bgtu.db0.impl

import ru.landgrafhomyak.bgtu.db0.storage.api.ModeratorPermissions
import ru.landgrafhomyak.bgtu.db0.storage.api.QueueSet
import ru.landgrafhomyak.bgtu.db0.storage.api.Storage
import ru.landgrafhomyak.bgtu.db0.telegram.api.CallbackQuery
import ru.landgrafhomyak.bgtu.db0.telegram.api.Message
import ru.landgrafhomyak.bgtu.db0.telegram.api.TelegramBotApi

@Suppress("FunctionName")
class Handlers(private val storage: Storage) {
	fun onTextMessage(api: TelegramBotApi, message: Message) {
		val (command, args) = Parsers.tryParseCommand(message.rawText) ?: return
		when (command) {
			"queue" -> this._createQueue(api, message, args)
			"qopen" -> this._openQueue(api, message, args)
			"qfinalize" -> this._closeQueue(api, message, args)
		}
	}

	fun onCallbackQuery(api: TelegramBotApi, query: CallbackQuery) {
		when (query.data) {
			"qpush" -> when (val result = this.storage.addUserToQueue(query.message.chatId, query.message.messageId, query.user.id, query.user.displayName)) {
				Storage.Result_AddOrRemoveUser.ALREADY_IN_STATE -> api.answerCallbackQuery(query, "")
				Storage.Result_AddOrRemoveUser.QUEUE_NOT_FOUND -> api.answerCallbackQuery(query, "\u26A0\uFE0F NullPointerException")
				Storage.Result_AddOrRemoveUser.QUEUE_OVERFLOW -> api.answerCallbackQuery(query, "\u26A0\uFE0F QueueOverflowException")
				Storage.Result_AddOrRemoveUser.QUEUE_IS_FINAL -> api.answerCallbackQuery(query, "\u26A0\uFE0F AccessToFinalQueueException")
				is Storage.Result_AddOrRemoveUser.SUCCESSFUL -> {
					api.answerCallbackQuery(query, "\u2705 Pushed")
					api.editMessageText(query.message, this.__fmtQueueMessage(result.updatedQueue))
				}
			}

			"qpop" -> when (val result = this.storage.removeUserFromQueue(query.message.chatId, query.message.messageId, query.user.id)) {
				Storage.Result_AddOrRemoveUser.QUEUE_NOT_FOUND -> api.answerCallbackQuery(query, "\u26A0\uFE0F NullPointerException")
				Storage.Result_AddOrRemoveUser.QUEUE_OVERFLOW -> api.answerCallbackQuery(query, "\u26A0\uFE0F QueueOverflowException")
				Storage.Result_AddOrRemoveUser.QUEUE_IS_FINAL -> api.answerCallbackQuery(query, "\u26A0\uFE0F AccessToFinalQueueException")
				Storage.Result_AddOrRemoveUser.ALREADY_IN_STATE -> api.answerCallbackQuery(query, "")
				is Storage.Result_AddOrRemoveUser.SUCCESSFUL -> {
					api.answerCallbackQuery(query, "\u2705 Popped")
					api.editMessageText(query.message, this.__fmtQueueMessage(result.updatedQueue))
				}
			}

			"qup" -> when (val result = this.storage.moveUserToStartOfQueue(query.message.chatId, query.message.messageId, query.user.id)) {
				Storage.Result_MoveToStartOrEnd.QUEUE_NOT_FOUND -> api.answerCallbackQuery(query, "\u26A0\uFE0F NullPointerException")
				Storage.Result_MoveToStartOrEnd.NOT_IN_QUEUE -> api.answerCallbackQuery(query, "\u26A0\uFE0F UserNotFoundException")
				Storage.Result_MoveToStartOrEnd.ON_EDGE -> api.answerCallbackQuery(query, "\u26A0\uFE0F Can't accelerate")
				Storage.Result_MoveToStartOrEnd.QUEUE_IS_FINAL -> api.answerCallbackQuery(query, "\u26A0\uFE0F AccessToFinalQueueException")
				is Storage.Result_MoveToStartOrEnd.SUCCESSFUL -> {
					api.answerCallbackQuery(query, "\u2705 Time goes faster...")
					api.editMessageText(query.message, this.__fmtQueueMessage(result.updatedQueue))
				}
			}

			"qdown" -> when (val result = this.storage.moveUserToEndOfQueue(query.message.chatId, query.message.messageId, query.user.id)) {
				Storage.Result_MoveToStartOrEnd.QUEUE_NOT_FOUND -> api.answerCallbackQuery(query, "\u26A0\uFE0F NullPointerException")
				Storage.Result_MoveToStartOrEnd.NOT_IN_QUEUE -> api.answerCallbackQuery(query, "\u26A0\uFE0F UserNotFoundException")
				Storage.Result_MoveToStartOrEnd.ON_EDGE -> api.answerCallbackQuery(query, "\u26A0\uFE0F Can't slow down")
				Storage.Result_MoveToStartOrEnd.QUEUE_IS_FINAL -> api.answerCallbackQuery(query, "\u26A0\uFE0F AccessToFinalQueueException")
				is Storage.Result_MoveToStartOrEnd.SUCCESSFUL -> {
					api.answerCallbackQuery(query, "\u2705 Time goes slower...")
					api.editMessageText(query.message, this.__fmtQueueMessage(result.updatedQueue))
				}
			}

			"qprev" -> when (val result = this.storage.moveUserToPrevQueue(query.message.chatId, query.message.messageId, query.user.id)) {
				Storage.Result_MoveToNextOrPrevQueue.QUEUE_NOT_FOUND -> api.answerCallbackQuery(query, "\u26A0\uFE0F NullPointerException")
				Storage.Result_MoveToNextOrPrevQueue.QUEUE_IS_FINAL -> api.answerCallbackQuery(query, "\u26A0\uFE0F AccessToFinalQueueException")
				Storage.Result_MoveToNextOrPrevQueue.NOT_IN_QUEUE -> api.answerCallbackQuery(query, "\u26A0\uFE0F UserNotFoundException")
				Storage.Result_MoveToNextOrPrevQueue.ON_EDGE -> api.answerCallbackQuery(query, "\u26A0\uFE0F You already at first queue")
				is Storage.Result_MoveToNextOrPrevQueue.SUCCESSFUL -> {
					api.answerCallbackQuery(query, "\u2705 Queue changed")
					api.editMessageText(query.message, this.__fmtQueueMessage(result.updatedQueue))
				}
			}

			"qnext" -> when (val result = this.storage.moveUserToNextQueue(query.message.chatId, query.message.messageId, query.user.id)) {
				Storage.Result_MoveToNextOrPrevQueue.QUEUE_NOT_FOUND -> api.answerCallbackQuery(query, "\u26A0\uFE0F NullPointerException")
				Storage.Result_MoveToNextOrPrevQueue.QUEUE_IS_FINAL -> api.answerCallbackQuery(query, "\u26A0\uFE0F AccessToFinalQueueException")
				Storage.Result_MoveToNextOrPrevQueue.NOT_IN_QUEUE -> api.answerCallbackQuery(query, "\u26A0\uFE0F UserNotFoundException")
				Storage.Result_MoveToNextOrPrevQueue.ON_EDGE -> api.answerCallbackQuery(query, "\u26A0\uFE0F You already at last queue")
				is Storage.Result_MoveToNextOrPrevQueue.SUCCESSFUL -> {
					api.answerCallbackQuery(query, "\u2705 Queue changed")
					api.editMessageText(query.message, this.__fmtQueueMessage(result.updatedQueue))
				}
			}
		}
	}


	private fun __checkPermissions(api: TelegramBotApi, chat: Long, user: Long, permission: ModeratorPermissions.PermissionType): Boolean =
true||		api.isChatOwner(chat, user) || this.storage.getPermissions(chat, user)?.get(permission) ?: false


	private fun _createQueue(api: TelegramBotApi, message: Message, commandArgs: String?) {
		if (!this.__checkPermissions(api, message.chatId, message.fromUser.id, ModeratorPermissions.PermissionType.CAN_CREATE_QUEUES)) {
			api.sendMessage(message, "<b>\u26A0\uFE0F У вас здесь нет власти чтобы создавать куеуе</b>")
			return
		}

		if (commandArgs.isNullOrBlank()) {
			api.sendMessage(message, "<b>\u26A0\uFE0F Сдесь могла быть ваша куеуе, но вы не указали ее название</b>")
			return
		}

		val names = commandArgs.split(Parsers.spaceSeparator)
		if (!names.all(Parsers::checkIsNameValid)) {
			api.sendMessage(message, "<b>\u26A0\uFE0F Сдесь могла быть ваша куеуе, но вы не умеете в названия переменных</b>")
			return
		}

		val setName = names.first()
		val queueNames = names.slice(1..<names.size)


		val queueMessage = api.sendMessage(
			message,
			this.__fmtQueueMessage(EmptyFakeQueueSet(setName, queueNames)),
			if (queueNames.isEmpty()) Keyboards.singleQueueKeyboard else Keyboards.multiQueueKeyboard
		)

		try {


			this.storage.allocQueue(message.chatId, queueMessage.messageId, setName, queueNames)
			return
		} catch (t: Throwable) {
			api.editMessageText(queueMessage, "<b>\u274C Не удалось создать куеуе в базе данных</b>", keepButtons = false)
			throw t
		}
	}

	private fun __fmtQueueMessage(queueSet: QueueSet): String = buildString {
		this@buildString.append("<b>").append(if (queueSet.isOpen) "open" else "final").append("</b>")

		this@buildString.append(" куеуе ")
		this@buildString.append("<u><b><i>").append(queueSet.name).append("</i></b></u>")
		this@buildString.append(" {\n")

		for (queue in queueSet.queues) {
			queue.name?.let { qname ->
				this@buildString.append("<u>").append(qname).append("</u>:\n")
			}
			queue.members.forEachIndexed { i, u ->
				if (u != null) {
					this@buildString.append("<code>  ").append(i).append(" </code>")
					this@buildString.append("<a href='tg://user?id=").append(u.telegramId).append("'>").append(u.displayName).append("</a>\n")
				} else {
					this@buildString.append("<code>  ").append(i).append("</code>\n")
				}
			}
		}
		this@buildString.append("}")
	}

	private fun _closeQueue(api: TelegramBotApi, message: Message, commandArgs: String?) {
		if (!this.__checkPermissions(api, message.chatId, message.fromUser.id, ModeratorPermissions.PermissionType.CAN_CLOSE_OR_OPEN_QUEUES)) {
			api.sendMessage(message, "<b>\u26A0\uFE0F AccessDeniedException</b>")
			return
		}

		if (!commandArgs.isNullOrBlank()) {
			api.sendMessage(message, "<b>\u26A0\uFE0F TypeError: Too many arguments</b>")
			return
		}

		val queueMessage = message.replyToMessage
		if (queueMessage == null) {
			api.sendMessage(message, "<b>\u26A0\uFE0F NullPointerException</b>")
			return
		}

		when (val result = this.storage.closeQueue(message.chatId, queueMessage.messageId)) {
			Storage.Result_QueueCloseOrOpen.ALREADY_IN_STATE -> {
				api.sendMessage(message, "<b>\u26A0\uFE0F You are too late, it's already final</b>")

			}

			Storage.Result_QueueCloseOrOpen.QUEUE_NOT_FOUND -> {
				api.sendMessage(message, "<b>\u26A0\uFE0F NullPointerException</b>")
				return
			}

			is Storage.Result_QueueCloseOrOpen.SUCCESSFUL -> {
				api.editMessageText(queueMessage, this.__fmtQueueMessage(result.updatedQueue))
				api.sendMessage(message, "<b>\u2705 Queue finalized</b>")
			}
		}
	}

	private fun _openQueue(api: TelegramBotApi, message: Message, commandArgs: String?) {
		if (!this.__checkPermissions(api, message.chatId, message.fromUser.id, ModeratorPermissions.PermissionType.CAN_CLOSE_OR_OPEN_QUEUES)) {
			api.sendMessage(message, "<b>\u26A0\uFE0F AccessDeniedException</b>")
			return
		}

		if (!commandArgs.isNullOrBlank()) {
			api.sendMessage(message, "<b>\u26A0\uFE0F TypeError: Too many arguments</b>")
			return
		}

		val queueMessage = message.replyToMessage
		if (queueMessage == null) {
			api.sendMessage(message, "<b>\u26A0\uFE0F NullPointerException</b>")
			return
		}

		when (val result = this.storage.reopenQueue(message.chatId, queueMessage.messageId)) {
			Storage.Result_QueueCloseOrOpen.ALREADY_IN_STATE -> {
				api.sendMessage(message, "<b>\u26A0\uFE0F You are so early, it's mutable yet</b>")

			}

			Storage.Result_QueueCloseOrOpen.QUEUE_NOT_FOUND -> {
				api.sendMessage(message, "<b>\u26A0\uFE0F NullPointerException</b>")
				return
			}

			is Storage.Result_QueueCloseOrOpen.SUCCESSFUL -> {
				api.editMessageText(queueMessage, this.__fmtQueueMessage(result.updatedQueue))
				api.sendMessage(message, "<b>\u2705 Queue finalized</b>")
			}
		}
	}
}