package ru.landgrafhomyak.bgtu.db0.storage.api

@Suppress("ConvertObjectToDataObject")
interface Storage {
	fun allocQueue(chat: Long, message: Long, setName: String, queueNames: List<String> = emptyList())

	@Suppress("ClassName")
	sealed class Result_QueueCloseOrOpen {
		class SUCCESSFUL(val updatedQueue: QueueSet) : Result_QueueCloseOrOpen()
		object QUEUE_NOT_FOUND : Result_QueueCloseOrOpen()
		object ALREADY_IN_STATE : Result_QueueCloseOrOpen()
	}

	fun closeQueue(chat: Long, message: Long): Result_QueueCloseOrOpen
	fun reopenQueue(chat: Long, message: Long): Result_QueueCloseOrOpen

	@Suppress("ClassName")
	sealed class Result_AddOrRemoveUser {
		class SUCCESSFUL(val updatedQueue: QueueSet) : Result_AddOrRemoveUser()
		object QUEUE_NOT_FOUND : Result_AddOrRemoveUser()
		object QUEUE_OVERFLOW : Result_AddOrRemoveUser()
		object ALREADY_IN_STATE : Result_AddOrRemoveUser()
		object QUEUE_IS_FINAL : Result_AddOrRemoveUser()
	}

	fun addUserToQueue(chat: Long, message: Long, user: Long, userDisplayName: String): Result_AddOrRemoveUser
	fun removeUserFromQueue(chat: Long, message: Long, user: Long): Result_AddOrRemoveUser


	@Suppress("ClassName")
	sealed class Result_MoveToNextOrPrevQueue {
		class SUCCESSFUL(val updatedQueue: QueueSet) : Result_MoveToNextOrPrevQueue()
		object QUEUE_NOT_FOUND : Result_MoveToNextOrPrevQueue()
		object NOT_IN_QUEUE : Result_MoveToNextOrPrevQueue()
		object ON_EDGE : Result_MoveToNextOrPrevQueue()
		object QUEUE_IS_FINAL : Result_MoveToNextOrPrevQueue()
	}

	fun moveUserToNextQueue(chat: Long, message: Long, user: Long): Result_MoveToNextOrPrevQueue
	fun moveUserToPrevQueue(chat: Long, message: Long, user: Long): Result_MoveToNextOrPrevQueue

	@Suppress("ClassName")
	sealed class Result_MoveToStartOrEnd {
		class SUCCESSFUL(val updatedQueue: QueueSet) : Result_MoveToStartOrEnd()
		object QUEUE_NOT_FOUND : Result_MoveToStartOrEnd()
		object NOT_IN_QUEUE : Result_MoveToStartOrEnd()
		object ON_EDGE : Result_MoveToStartOrEnd()
		object QUEUE_IS_FINAL : Result_MoveToStartOrEnd()
	}

	fun moveUserToStartOfQueue(chat: Long, message: Long, user: Long): Result_MoveToStartOrEnd
	fun moveUserToEndOfQueue(chat: Long, message: Long, user: Long): Result_MoveToStartOrEnd

	/**
	 * @return `false` on success, `true` if already moderator
	 */
	fun addModerator(chat: Long, user: Long): Boolean

	/**
	 * @return `false` on success, `true` if already not moderator
	 */
	fun removeModerator(chat: Long, message: Long): Boolean

	fun getPermissions(chat: Long, user: Long): ModeratorPermissions?

	@Suppress("ClassName")
	sealed class Result_SetModeratorPermission {
		class SUCCESSFUL(val newPermissions: QueueSet) : Result_SetModeratorPermission()
		object NOT_MODERATOR : Result_SetModeratorPermission()
		object ALREADY_SET : Result_SetModeratorPermission()
	}

	fun setPermission(chat: Long, user: Long, permissionType: ModeratorPermissions.PermissionType, value: Boolean): Result_SetModeratorPermission
}